pub trait ReadEntity: DeserializeOwned + Select { }Expand description
Supertrait for entities that can be retrieved from a Microsoft Dataverse environment
This should be implemented by data structures you want to use with
the following functions in Client:
retrieve(...)retrieve_multiple(...)
Examples
use serde::Deserialize;
use uuid::Uuid;
use powerplatform_dataverse_service_client::{
entity::ReadEntity,
select::Select
};
#[derive(Deserialize)]
struct Contact {
contactid: Uuid,
firstname: String,
lastname: String,
}
impl ReadEntity for Contact {}
impl Select for Contact {
fn get_columns() -> &'static [&'static str] {
&["contactid", "firstname", "lastname"]
}
}Dyn Compatibility
This trait is not dyn compatible.
In older versions of Rust, dyn compatibility was called "object safety", so this trait is not object safe.